{-# LANGUAGE CPP, FlexibleContexts, FlexibleInstances, MultiParamTypeClasses, ScopedTypeVariables, TypeSynonymInstances #-}
{-# OPTIONS_GHC -Wall #-}
module System.Process.Read.InterleavedOld
    ( NonBlocking(..)
    , readInterleaved
    ) where

import Control.Applicative ((<$>))
import Control.Concurrent (threadDelay)
import Control.Exception as E (catch)
import qualified Data.ByteString as B
import qualified Data.ByteString.Lazy as L
import Data.ListLike (ListLike(length))
import Data.Maybe (isNothing)
import Data.Monoid (Monoid, mempty, (<>), mconcat)
import Data.Word (Word8)
import qualified GHC.IO.Exception as E
import Prelude hiding (null, length, rem)
import System.IO hiding (hPutStr, hGetContents)
import System.Process.Read (ListLikePlus(..))
import System.IO.Unsafe (unsafeInterleaveIO)

-- | Class of types which can also be used by 'System.Process.Read.readProcessChunks'.
class ListLikePlus a c => NonBlocking a c where
  hGetSome :: Handle -> LengthType a -> IO a
  toChunks :: a -> [a]

{- Its not safe to read a bytestring chunk and then convert it to a string, the chunk
   might end with part of a character.
instance NonBlocking String Char where
  -- hGetNonBlocking n h = (toString . B.concat . L.toChunks) <$> L.hGetNonBlocking n h
  hGetSome h n = toList <$> B.hGetSome h n
  toChunks = error "toChunks"
-}

instance NonBlocking B.ByteString Word8 where
  -- hGetNonBlocking = B.hGetNonBlocking
  hGetSome = B.hGetSome
  toChunks = (: [])

instance NonBlocking L.ByteString Word8 where
  -- hGetNonBlocking = L.hGetNonBlocking
  hGetSome h n = (L.fromChunks . (: [])) <$> B.hGetSome h (fromIntegral n)
  toChunks = Prelude.map (L.fromChunks . (: [])) . L.toChunks

-- | Read input from a list of handles, using the function supplied
-- with each handle to convert the input into some Monoid.
readInterleaved :: (NonBlocking a c, Monoid b) => [(a -> b, Maybe Handle)] -> IO b
readInterleaved pairs | all (isNothing . snd) pairs = return mempty
readInterleaved pairs = do
  -- The available output has been processed, send input and read
  -- from the ready handles
  (handles', input) <- readReady uSecs pairs
  more <- unsafeInterleaveIO (readInterleaved (zip (map fst pairs) handles'))
  return $ input <> more

uSecs :: Int
uSecs = 8		-- minimum wait time, doubles each time nothing is ready

data Readyness = Ready | Unready | Closed deriving (Eq)

-- | Wait until at least one handle is ready and then read output.  If
-- none of the output descriptors are ready for reading the function
-- sleeps and tries again.
readReady :: (NonBlocking a c, Monoid b) => Int -> [(a -> b, Maybe Handle)] -> IO ([Maybe Handle], b)
readReady waitUSecs pairs = do
  readys <- mapM (maybe (return Closed) readyness) (map snd pairs)
  case () of
    _ | all (== Closed) readys -> return (map (const Nothing) pairs, mempty)
    _ | all (/= Ready) readys ->
          -- Input handle closed and there are no ready output handles,
          -- wait a bit
          do threadDelay waitUSecs
             readReady (min maxUSecs (2 * waitUSecs)) pairs
    _ -> -- One or both output handles are ready, try to read from
         -- them.  If (out1 ++ out2) is an empty list we know that
         -- we have reached EOF on both descriptors, because at
         -- least one was ready and nextOut only returns [] for a
         -- ready file descriptor on EOF.
         do (handles, inputs) <- unzip <$> mapM (\ ((f, h), r) -> readSome h r f) (zip pairs readys)
            return $ (handles, mconcat inputs)

maxUSecs :: Int
maxUSecs = 100000	-- maximum wait time (microseconds)

-- | Test a handle's readyness.  This wrapper around hReady is
-- necessary because if we run hReady on a closed handle it gives
-- us an infinite number of EOF exceptions (last time I tried.)
readyness :: Handle -> IO Readyness
readyness h =
    hIsClosed h >>= checkReady
    where
      checkReady True = return Closed
      checkReady False =
          (hReady h >>= (\ flag -> return (if flag then Ready else Unready)))
            `catch` endOfFile (\ _ -> return Closed)
      endOfFile :: (IOError -> IO a) -> IOError -> IO a
      endOfFile eeof e = if E.ioe_type e == E.EOF then eeof e else ioError e

-- | Return the next input element and the updated handle from a
-- handle which is assumed ready.  If the handle is closed or unready,
-- or we reach end of file, an empty input element is returned.
readSome :: (NonBlocking a c, Monoid b) => Maybe Handle -> Readyness -> (a -> b) -> IO (Maybe Handle, b)
readSome Nothing _ _ = return (Nothing, mempty)
readSome (Just h) Ready constructor =	-- Perform a read
   do -- We can call hGetSome because we know this handle is ready for reading.
      a <- hGetSome h (fromIntegral bufSize)
      case length a of
        -- A zero length read, unlike a zero length write, always
        -- means EOF.
        0 -> do hClose h
                return (Nothing, mempty)
        -- Got some output
        _n -> return (Just h, constructor a)
readSome h _ _ = return (h, mempty)	-- Handle is closed or not ready

bufSize :: Int
bufSize = 4096		-- maximum chunk size
