-- | The 'readProcessChunks' function is a process reader that returns
-- a list (stream) of 'Output', which represent chunks of text read
-- from 'Stdout', 'Stderr', a 'Result' code, or an 'Exception'.  This
-- has the advantage of preserving the order in which these things
-- appeared.  The 'foldOutput', 'foldOutputsL', and 'foldOutputsR'
-- functions can be used to process the output stream.  The output
-- text can be any 'NonBlocking' instance, which needs to be implemented
-- using a non-blocking read function like 'B.hGetSome'.

{-# LANGUAGE CPP, FlexibleContexts, FlexibleInstances, MultiParamTypeClasses, ScopedTypeVariables, TypeSynonymInstances #-}
{-# OPTIONS_GHC -Wall #-}
module System.Process.Read.ProcessOld
    ( readProcessInterleaved
    ) where

import Control.Concurrent (forkIO, newEmptyMVar, putMVar, takeMVar)
import Control.Exception as E (onException, catch, mask, try, throwIO, SomeException)
import Control.Monad (unless)
import Data.ListLike (ListLike(..), ListLikeIO(..))
import Data.Monoid (Monoid, (<>))
import GHC.IO.Exception (IOErrorType(ResourceVanished), IOException(ioe_type))
import Prelude hiding (null, length, rem)
import System.Exit (ExitCode)
import System.IO hiding (hPutStr, hGetContents)
import System.Process.Read (ListLikePlus(..))
import System.Process.Read.Interleaved (Chunked)
import System.Process.Read.InterleavedOld (readInterleaved)
import System.Process (CreateProcess(..), StdStream(CreatePipe), ProcessHandle,
                       createProcess, waitForProcess, terminateProcess)

-- | This is a process runner which (at the cost of a busy loop) gives
-- you the chunks of text read from stdout and stderr interleaved in
-- the order they were written, along with any ResourceVanished
-- exxception that might occur.  Its interface is similar to
-- 'System.Process.Read.readModifiedProcessWith', though the
-- implementation is somewhat alarming.
readProcessInterleaved :: (ListLikePlus a c, Monoid b) =>
                          (ExitCode -> b) -> (a -> b) -> (a -> b)
                       -> CreateProcess -> a -> IO b
readProcessInterleaved codef outf errf p input = mask $ \ restore -> do
  (Just inh, Just outh, Just errh, pid) <-
      createProcess (p {std_in = CreatePipe, std_out = CreatePipe, std_err = CreatePipe })

  binary input [inh, outh, errh]

  flip onException
    (do hClose inh; hClose outh; hClose errh;
        terminateProcess pid; waitForProcess pid) $ restore $ do

    waitOut <- forkWait $ elements codef outf errf pid outh errh

    -- now write and flush any input
    (do unless (null input) (hPutStr inh input >> hFlush inh)
        hClose inh) `E.catch` resourceVanished (\ _e -> return ())

    -- wait on the output
    waitOut

-- | Take the info returned by 'createProcess' and gather and return
-- the stdout and stderr of the process.
elements :: (Chunked a c, Monoid b) =>
            (ExitCode -> b) -> (a -> b) -> (a -> b)
         -> ProcessHandle -> Handle -> Handle -> IO b
elements codef outf errf pid outh errh = do
  output <- readInterleaved [(outf, Just outh), (errf, Just errh)]
  -- EOF on both output descriptors, get exit code.  It can be
  -- argued that the result will always contain exactly one exit
  -- code if traversed to its end, because the only case of
  -- 'elements' that does not recurse is the one that adds a Result,
  -- and there is nowhere else where a Result is added.  However,
  -- the process doing the traversing may die before that end is
  -- reached.
  result <- waitForProcess pid
  return (output <> codef result)

forkWait :: IO a -> IO (IO a)
forkWait a = do
  res <- newEmptyMVar
  _ <- mask $ \restore -> forkIO $ try (restore a) >>= putMVar res
  return (takeMVar res >>= either (\ex -> throwIO (ex :: SomeException)) return)

-- | Wrapper for a process that provides a handler for the
-- ResourceVanished exception.  This is frequently an exception we
-- wish to ignore, because many processes will deliberately exit
-- before they have read all of their input.
resourceVanished :: (IOError -> IO a) -> IOError -> IO a
resourceVanished epipe e = if ioe_type e == ResourceVanished then epipe e else ioError e
